#!/usr/bin/env node

'use strict';

const path = require('path');
const mri = require('mri');
const listhen = require('listhen');
const h3 = require('h3');
const fs = require('fs');
const chokidar = require('chokidar');
const anymatch = require('anymatch');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

const mri__default = /*#__PURE__*/_interopDefaultLegacy(mri);
const anymatch__default = /*#__PURE__*/_interopDefaultLegacy(anymatch);

const suspectProtoRx = /"(?:_|\\u005[Ff])(?:_|\\u005[Ff])(?:p|\\u0070)(?:r|\\u0072)(?:o|\\u006[Ff])(?:t|\\u0074)(?:o|\\u006[Ff])(?:_|\\u005[Ff])(?:_|\\u005[Ff])"\s*:/;
const suspectConstructorRx = /"(?:c|\\u0063)(?:o|\\u006[Ff])(?:n|\\u006[Ee])(?:s|\\u0073)(?:t|\\u0074)(?:r|\\u0072)(?:u|\\u0075)(?:c|\\u0063)(?:t|\\u0074)(?:o|\\u006[Ff])(?:r|\\u0072)"\s*:/;
const JsonSigRx = /^["{[]|^-?[0-9][0-9.]{0,14}$/;
function jsonParseTransform(key, value) {
  if (key === "__proto__" || key === "constructor") {
    return;
  }
  return value;
}
function destr(val) {
  if (typeof val !== "string") {
    return val;
  }
  const _lval = val.toLowerCase();
  if (_lval === "true") {
    return true;
  }
  if (_lval === "false") {
    return false;
  }
  if (_lval === "null") {
    return null;
  }
  if (_lval === "nan") {
    return NaN;
  }
  if (_lval === "infinity") {
    return Infinity;
  }
  if (_lval === "undefined") {
    return void 0;
  }
  if (!JsonSigRx.test(val)) {
    return val;
  }
  try {
    if (suspectProtoRx.test(val) || suspectConstructorRx.test(val)) {
      return JSON.parse(val, jsonParseTransform);
    }
    return JSON.parse(val);
  } catch (_e) {
    return val;
  }
}

function defineDriver(factory) {
  return factory;
}

const memory = defineDriver(() => {
  const data = new Map();
  return {
    hasItem(key) {
      return data.has(key);
    },
    getItem(key) {
      return data.get(key) || null;
    },
    setItem(key, value) {
      data.set(key, value);
    },
    removeItem(key) {
      data.delete(key);
    },
    getKeys() {
      return Array.from(data.keys());
    },
    clear() {
      data.clear();
    },
    dispose() {
      data.clear();
    }
  };
});

function normalizeKey(key) {
  if (!key) {
    return "";
  }
  return key.replace(/[/\\]/g, ":").replace(/^:|:$/g, "");
}
function normalizeBase(base) {
  base = normalizeKey(base);
  return base ? base + ":" : "";
}
function wrapToPromise(val) {
  if (!val || typeof val.then !== "function") {
    return Promise.resolve(val);
  }
  return val;
}
function asyncCall(fn, ...args) {
  try {
    return wrapToPromise(fn(...args));
  } catch (err) {
    return Promise.reject(err);
  }
}
function isPrimitive(arg) {
  const type = typeof arg;
  return arg === null || type !== "object" && type !== "function";
}
function stringify(arg) {
  return isPrimitive(arg) ? arg + "" : JSON.stringify(arg);
}

function createStorage(opts = {}) {
  const ctx = {
    mounts: {"": opts.driver || memory()},
    mountpoints: [""],
    watching: false,
    watchListeners: []
  };
  const getMount = (key) => {
    for (const base of ctx.mountpoints) {
      if (key.startsWith(base)) {
        return {
          relativeKey: key.substr(base.length),
          driver: ctx.mounts[base]
        };
      }
    }
    return {
      relativeKey: key,
      driver: ctx.mounts[""]
    };
  };
  const getMounts = (base) => {
    return ctx.mountpoints.filter((mountpoint) => base.length < mountpoint.length || base.startsWith(mountpoint)).map((mountpoint) => ({
      mountpoint,
      driver: ctx.mounts[mountpoint]
    }));
  };
  const onChange = (event, key) => {
    if (!ctx.watching) {
      return;
    }
    key = normalizeKey(key);
    for (const listener of ctx.watchListeners) {
      listener(event, key);
    }
  };
  const startWatch = async () => {
    if (ctx.watching) {
      return;
    }
    ctx.watching = true;
    for (const mountpoint in ctx.mounts) {
      await watch(ctx.mounts[mountpoint], onChange, mountpoint);
    }
  };
  const storage = {
    hasItem(key) {
      key = normalizeKey(key);
      const {relativeKey, driver} = getMount(key);
      return asyncCall(driver.hasItem, relativeKey);
    },
    getItem(key) {
      key = normalizeKey(key);
      const {relativeKey, driver} = getMount(key);
      return asyncCall(driver.getItem, relativeKey).then((val) => destr(val));
    },
    async setItem(key, value) {
      if (value === void 0) {
        return storage.removeItem(key);
      }
      key = normalizeKey(key);
      const {relativeKey, driver} = getMount(key);
      await asyncCall(driver.setItem, relativeKey, stringify(value));
      if (!driver.watch) {
        onChange("update", key);
      }
    },
    async removeItem(key) {
      key = normalizeKey(key);
      const {relativeKey, driver} = getMount(key);
      await asyncCall(driver.removeItem, relativeKey);
      if (!driver.watch) {
        onChange("remove", key);
      }
    },
    async getKeys(base) {
      base = normalizeBase(base);
      const keyGroups = await Promise.all(getMounts(base).map(async (mount) => {
        const rawKeys = await asyncCall(mount.driver.getKeys);
        return rawKeys.map((key) => mount.mountpoint + normalizeKey(key));
      }));
      const keys = keyGroups.flat();
      return base ? keys.filter((key) => key.startsWith(base)) : keys;
    },
    async clear(base) {
      base = normalizeBase(base);
      await Promise.all(getMounts(base).map((m) => asyncCall(m.driver.clear)));
    },
    async dispose() {
      await Promise.all(Object.values(ctx.mounts).map((driver) => dispose(driver)));
    },
    mount(base, driver) {
      base = normalizeBase(base);
      if (base && ctx.mounts[base]) {
        throw new Error(`already mounted at ${base}`);
      }
      if (base) {
        ctx.mountpoints.push(base);
        ctx.mountpoints.sort((a, b) => b.length - a.length);
      }
      ctx.mounts[base] = driver;
      if (ctx.watching) {
        Promise.resolve(watch(driver, onChange, base)).catch(console.error);
      }
      return storage;
    },
    async unmount(base, _dispose = true) {
      base = normalizeBase(base);
      if (!base || !ctx.mounts[base]) {
        return;
      }
      if (_dispose) {
        await dispose(ctx.mounts[base]);
      }
      ctx.mountpoints = ctx.mountpoints.filter((key) => key !== base);
      delete ctx.mounts[base];
    },
    async watch(callback) {
      await startWatch();
      ctx.watchListeners.push(callback);
    }
  };
  return storage;
}
function watch(driver, onChange, base) {
  if (driver.watch) {
    return driver.watch((event, key) => onChange(event, base + key));
  }
}
async function dispose(driver) {
  if (typeof driver.dispose === "function") {
    await asyncCall(driver.dispose);
  }
}

function createStorageServer(storage, _opts = {}) {
  const app = h3.createApp();
  app.useAsync(async (req, res) => {
    if (req.method === "GET") {
      const val = await storage.getItem(req.url);
      if (!val) {
        const keys = await storage.getKeys(req.url);
        return keys.map((key) => key.replace(/:/g, "/"));
      }
      return stringify(val);
    }
    if (req.method === "HEAD") {
      const _hasItem = await storage.hasItem(req.url);
      res.statusCode = _hasItem ? 200 : 404;
      return "";
    }
    if (req.method === "PUT") {
      const val = await h3.useBody(req);
      await storage.setItem(req.url, val);
      return "OK";
    }
    if (req.method === "DELETE") {
      await storage.removeItem(req.url);
      return "OK";
    }
    throw h3.createError({
      statusCode: 405,
      statusMessage: "Method Not Allowd"
    });
  });
  return {
    handle: app
  };
}

function ignoreNotfound(err) {
  return err.code === "ENOENT" || err.code === "EISDIR" ? null : err;
}
function ignoreExists(err) {
  return err.code === "EEXIST" ? null : err;
}
async function writeFile(path$1, data) {
  await ensuredir(path.dirname(path$1));
  return fs.promises.writeFile(path$1, data, "utf8");
}
function readFile(path) {
  return fs.promises.readFile(path, "utf8").catch(ignoreNotfound);
}
function unlink(path) {
  return fs.promises.unlink(path).catch(ignoreNotfound);
}
function readdir(dir) {
  return fs.promises.readdir(dir, {withFileTypes: true}).catch(ignoreNotfound).then((r) => r || []);
}
async function ensuredir(dir) {
  if (fs.existsSync(dir)) {
    return;
  }
  await ensuredir(path.dirname(dir)).catch(ignoreExists);
  await fs.promises.mkdir(dir).catch(ignoreExists);
}
async function readdirRecursive(dir, ignore) {
  if (ignore && ignore(dir)) {
    return [];
  }
  const entries = await readdir(dir);
  const files = [];
  await Promise.all(entries.map(async (entry) => {
    const entryPath = path.resolve(dir, entry.name);
    if (entry.isDirectory()) {
      const dirFiles = await readdirRecursive(entryPath, ignore);
      files.push(...dirFiles.map((f) => entry.name + "/" + f));
    } else {
      if (ignore && !ignore(entry.name)) {
        files.push(entry.name);
      }
    }
  }));
  return files;
}
async function rmRecursive(dir) {
  const entries = await readdir(dir);
  await Promise.all(entries.map((entry) => {
    const entryPath = path.resolve(dir, entry.name);
    if (entry.isDirectory()) {
      return rmRecursive(entryPath).then(() => fs.promises.rmdir(entryPath));
    } else {
      return fs.promises.unlink(entryPath);
    }
  }));
}

const fsDriver = defineDriver((opts = {}) => {
  if (!opts.base) {
    throw new Error("base is required");
  }
  if (!opts.ignore) {
    opts.ignore = [
      "**/node_modules/**",
      "**/.git/**"
    ];
  }
  opts.base = path.resolve(opts.base);
  const r = (key) => path.join(opts.base, key.replace(/:/g, "/"));
  let _watcher;
  return {
    hasItem(key) {
      return fs.existsSync(r(key));
    },
    getItem(key) {
      return readFile(r(key));
    },
    setItem(key, value) {
      return writeFile(r(key), value);
    },
    removeItem(key) {
      return unlink(r(key));
    },
    getKeys() {
      return readdirRecursive(r("."), anymatch__default['default'](opts.ignore || []));
    },
    async clear() {
      await rmRecursive(r("."));
    },
    async dispose() {
      if (_watcher) {
        await _watcher.close();
      }
    },
    watch(callback) {
      if (_watcher) {
        return;
      }
      return new Promise((resolve2, reject) => {
        _watcher = chokidar.watch(opts.base, {
          ignoreInitial: true,
          ignored: opts.ignore,
          ...opts.watchOptions
        }).on("ready", resolve2).on("error", reject).on("all", (eventName, path$1) => {
          path$1 = path.relative(opts.base, path$1);
          if (eventName === "change" || eventName === "add") {
            callback("update", path$1);
          } else if (eventName === "unlink") {
            callback("remove", path$1);
          }
        });
      });
    }
  };
});

async function main() {
  const args = mri__default['default'](process.argv.splice(2));
  if (args.help) {
    console.log("Usage: npx unstorage [rootDir]");
    process.exit(0);
  }
  const rootDir = path.resolve(args._[0] || ".");
  const storage = createStorage({
    driver: fsDriver({base: rootDir})
  });
  const storageServer = createStorageServer(storage);
  await listhen.listen(storageServer.handle, {
    name: "Storage server",
    port: 8080
  });
}
main().catch((err) => {
  console.error(err);
  process.exit(1);
});
